import os
import pandas as pd

def complement(seq):
	complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', \
	'M': 'K', 'K': 'M', 'R': 'Y', 'Y': 'R', 'W': 'W', 'S': 'S', 'V': 'B', 'B': 'V', 'H': 'D', 'D': 'H'}
	bases = list(seq)
	bases = [complement[base] for base in bases]
	return ''.join(bases)


def ParseSequence(FastaFile,CoList,Strand):

	Scaf = CoList[0][0];ChromSeq = '';Flag = 0
	fin = open(FastaFile,'r')
	for line in fin:
		if '>' in line[0]:
			SeqId = line.strip().split()[0].replace('>','')
			if SeqId == Scaf:
				Flag = 1
			else:
				Flag = 0
		else:
			if Flag == 1:
				ChromSeq += line.strip().upper()
	fin.close()

	Seq = ''
	if Strand == '+':
		for Scaf,Start,End in CoList:
			Seq += ChromSeq[Start-1:End]

	else:
		for Scaf,Start,End in CoList:
			Seq += complement(ChromSeq[Start-1:End])[::-1]

	return Seq


def ParseGeneId(FileName):

	fin = open(FileName,'r');IdList = {}
	for line in fin:

		if '>' in line:
			SeqId = line.strip().split(', ')[1].split()[0]
			IdList.setdefault(SeqId,'')

	fin.close()

	return IdList


def MakeCoDf_n_CdsEndDic(IdList,GffFile):

	TransCoDic = {};CdsEndDic = {}
	fin = open(GffFile,'r')
	for line in fin:

		if not '#' in line[0]:

			Type = line.strip().split('\t')[2]
			Strand = line.strip().split('\t')[6]

			if Type == 'gene':

				GeneId = line.strip().split('ID=gene-')[1].split(';')[0]

			elif Type in ['mRNA','transcript']:

				GeneId = line.strip().split(';gene=')[1].split(';')[0]
				TransId = line.strip().split('ID=rna-')[1].split(';')[0]

				if TransId.split('.')[0] in IdList:
					TransCoDic.setdefault((GeneId,TransId,Strand),[])

			elif Type == 'exon':

				if ';gene=' in line:
					GeneId = line.strip().split(';gene=')[1].split(';')[0]
				else:
					GeneId = line.strip().split(';Parent=rna-')[1].split(';')[0]

				if ';Parent=rna-' in line:
					TransId = line.strip().split(';Parent=rna-')[1].split(';')[0]
				else:
					TransId = line.strip().split(';Parent=')[1].split(';')[0]

				if TransId.split('.')[0] in IdList:

					ExonId = line.strip().split('ID=exon-')[1].split(';')[0]
					Scaf = line.strip().split('\t')[0]
					Start = int(line.strip().split('\t')[3])
					End = int(line.strip().split('\t')[4])
					TransCoDic[(GeneId,TransId,Strand)].append([Scaf,Start,End])

			elif Type == 'CDS':

				GeneId = line.strip().split(';gene=')[1].split(';')[0]
				if ';Parent=rna-' in line:
					TransId = line.strip().split(';Parent=rna-')[1].split(';')[0]
				else:
					TransId = line.strip().split(';Parent=')[1].split(';')[0]

				if Strand == '+':CdsEnd = int(line.strip().split('\t')[4])
				elif Strand == '-':CdsEnd = int(line.strip().split('\t')[3])

				if TransId.split('.')[0] in IdList:
					CdsEndDic[(GeneId,TransId)] = [Scaf,CdsEnd]

	return TransCoDic,CdsEndDic


def MakeUtrCoDf(IdList,GffFile,FastaFile):

	TransCoDic,CdsEndDic = MakeCoDf_n_CdsEndDic(IdList,GffFile)

	UtrCoDic = {}
	for GeneId,TransId,Strand in sorted(TransCoDic.keys()):

		UtrCoDic.setdefault((GeneId,TransId,Strand),[])

		if (GeneId,TransId) in CdsEndDic:

			Scaf,CdsEnd = CdsEndDic[(GeneId,TransId)]

			for Scaf,Start,End in TransCoDic[(GeneId,TransId,Strand)]:

				if Strand == '-':
					if Start < CdsEnd:
						if End < CdsEnd:
							UtrCoDic[(GeneId,TransId,Strand)].append([Scaf,Start,End])
						elif End >= CdsEnd:
							UtrCoDic[(GeneId,TransId,Strand)].append([Scaf,Start,CdsEnd-1])

				elif Strand == '+':
					if End > CdsEnd:
						if Start > CdsEnd:
							UtrCoDic[(GeneId,TransId,Strand)].append([Scaf,Start,End])
						elif Start <= CdsEnd:
							UtrCoDic[(GeneId,TransId,Strand)].append([Scaf,CdsEnd+1,End])

	UtrCoDf = {};Header = ['Gene','Transcript','Strand','Scaf','Start','End']
	for ColName in Header:UtrCoDf.setdefault(ColName,[])

	for GeneId,TransId,Strand in sorted(UtrCoDic.keys()):
		for Scaf,Start,End in UtrCoDic[(GeneId,TransId,Strand)]:
			Elements = [GeneId,TransId,Strand,Scaf,Start,End]
			for i in range(len(Elements)):UtrCoDf[Header[i]].append(Elements[i])

	UtrCoDf = pd.DataFrame(UtrCoDf,columns=Header)
	UtrCoDf.to_csv(UtrDfFile,sep='\t')	

	return UtrCoDic


def MakeFastaFile(UtrFastaFile,UtrCoDic):

	fout = open(UtrFastaFile,'w');SeqDic = {}
	for GeneId,TransId,Strand in sorted(UtrCoDic.keys()):
		if len(UtrCoDic[(GeneId,TransId,Strand)]) > 0:
			print('Parsing Sequence of %s %s %s ' % (GeneId,TransId,Strand))
			fout.write('>%s %s %s\n' % (GeneId,TransId,Strand))
			Seq = ParseSequence(FastaFile,UtrCoDic[(GeneId,TransId,Strand)],Strand)
			fout.write(Seq+'\n')
			SeqDic[(GeneId,TransId,Strand)] = Seq
	fout.close()

	return SeqDic


def CompareWithInputSeq(SeqDic):

	fin = open('cDAZL_motifs.txt','r');SeqDic2 = {}
	for line in fin:
		if '>' in line[0]:
			SeqId = line.strip().split(', ')[1]
			SeqDic2.setdefault(SeqId,'')
		else:
			SeqDic2[SeqId] += line.strip()
	fin.close()

	for GeneId,TransId,Strand in sorted(SeqDic.keys()):
		Seq1 = SeqDic[(GeneId,TransId,Strand)]
		Seq2 = SeqDic2[TransId.split('.')[0]].replace('U','T')
		if Seq1 == Seq2:
			print(GeneId,TransId,Strand)
		

def CheckForBindingSite(SeqDic,BsCntFile):

	BindingSite5 = ['TGTTT','TGTTA']
	BindingSite4 = ['TGTT','GTTT','GTTA','GTTG','GTTC']
	BindingSite7 = ['TTTCTTT','TTTGTTT']

	BsList = BindingSite5+BindingSite4+BindingSite7
	CntDic = {}
	for SeqId in sorted(SeqDic.keys()):
		
		for Bs in BsList:CntDic.setdefault((SeqId,Bs),0)

		for i in range(len(SeqDic[SeqId])-4):

			if SeqDic[SeqId][i:i+5] in BindingSite5:
				CntDic[(SeqId,SeqDic[SeqId][i:i+5])] += 1

		for i in range(len(SeqDic[SeqId])-3):
			
			if SeqDic[SeqId][i:i+4] in BindingSite4:
				CntDic[(SeqId,SeqDic[SeqId][i:i+4])] += 1

		for i in range(len(SeqDic[SeqId])-6):

			if SeqDic[SeqId][i:i+7] in BindingSite7:
				CntDic[(SeqId,SeqDic[SeqId][i:i+7])] += 1

	CntDf = {};Header = ['Gene','Transcript']+BsList
	for ColName in Header:CntDf.setdefault(ColName,[])

	for SeqId in sorted(SeqDic.keys()):
		GeneId,TransId,Strand = SeqId

		CntDf[Header[0]].append(GeneId)
		CntDf[Header[1]].append(TransId)
		for i in range(len(BsList)):
			CntDf[Header[2+i]].append(CntDic[(SeqId,BsList[i])])

	#for i in CntDf:
	#	print(i, len(CntDf[i]))

	CntDf = pd.DataFrame(CntDf,columns=Header)
	CntDf.to_csv(BsCntFile,sep='\t')


if __name__ == '__main__':

	IdList = {}
	fin = open('TranscriptList.txt','r')
	for line in fin:
		IdList.setdefault(line.strip(),'')
	fin.close()
	
	## Inputs ##
	GffFile = 'GCF_000002315.6_GRCg6a_genomic.gff'
	FastaFile = 'GCF_000002315.6_GRCg6a_genomic.fna'
	UtrDfFile = '3UtrDf.txt'
	UtrFastaFile = '3Utr.fa'

	## Output ##
	BsCntFile = 'Motifs3Utr.txt'

	if not UtrDfFile in os.listdir('./'):
		UtrCoDic = MakeUtrCoDf(IdList,GffFile,FastaFile)
	else:
		fin = open(UtrDfFile,'r');UtrCoDic = {}
		fin.readline()
		for line in fin:
			Ind,GeneId,TransId,Strand,Scaf,Start,End = line.strip().split('\t')
			UtrCoDic.setdefault((GeneId,TransId,Strand),[])
			UtrCoDic[(GeneId,TransId,Strand)].append([Scaf,int(Start),int(End)])
		fin.close()

	if not UtrFastaFile in os.listdir('./'):
		SeqDic = MakeFastaFile(UtrFastaFile,UtrCoDic)
	else:
		fin = open(UtrFastaFile,'r');SeqDic = {}
		for line in fin:
			if '>' in line[0]:
				GeneId,TransId,Strand = line.strip().replace('>','').split()
				SeqDic.setdefault((GeneId,TransId,Strand),'')
			else:
				SeqDic[(GeneId,TransId,Strand)] += line.strip()
		fin.close()
	
	CompareWithInputSeq(SeqDic)
	CheckForBindingSite(SeqDic,BsCntFile)

